/**
 * @file       sdCard.c
 * @brief      read/write in a sdcard throught SPI
 * 
 * @author     Laurent Saint-Marcel (lstmarcel@yahoo.fr)
 * @date       2009/06/26
 */

/*
gravement modifié par Silicium628  ;)
2015-07-03
*/

#include <avr/io.h>

#include "sdCard_628.h"
#include "sdCardProtocol.h"

// SPI port PIN configuration 
// The following configuration is valid for an ATMEGA324

#define SPI_PORT PORTB
#define SPI_DDR  DDRB
#define SPI_SS	    4	
#define SPI_MOSI		5
#define SPI_MISO		6	
#define SPI_SCK	    7	

// ----------------------------------------------------------------------------
// PRIVATE FUNCTIONS
// ----------------------------------------------------------------------------
#define WAIT_TIMEOUT   0xC350 // 16 bits timeout


uint16_t SDCARD_error;


// ---------------------------------------------------------------------------
// Set bit in IO port.
// ---------------------------------------------------------------------------
#define sbi(port, bit) (port) |= (1 << (bit))

// ---------------------------------------------------------------------------
// Clear bit in IO port.
// ---------------------------------------------------------------------------
#define cbi(port, bit) (port) &= ~(1 << (bit))

// ---------------------------------------------------------------------------
// set chip select to low (sdCard is selected is selected)
// ---------------------------------------------------------------------------
void sdSelectChip()
{
	cbi(SPI_PORT, SPI_SS);	
}

// ---------------------------------------------------------------------------
// set chip select to low (sdCard is selected is selected)
// ---------------------------------------------------------------------------
void sdDeSelectChip()
{
	sbi(SPI_PORT, SPI_SS);	
}

// ---------------------------------------------------------------------------
// set slow clock speed during card initialization
// ---------------------------------------------------------------------------
void sdSetSlowClockSpeed()
{
	// SPI enabled
	// master mode
	// SPIclock = Fclock/64=124kHz (less than 4OOkHz for Multimedia cards)
  SPCR = (1 << SPE) | (1 << MSTR) | (1 << SPR1) | (1 << SPR0);   // p:134 du pdf
	sbi(SPSR, SPI2X);   // p:134
}

// ---------------------------------------------------------------------------
// set fast SPI clock speed for card read/write
// ---------------------------------------------------------------------------
void sdSetFastClockSpeed()
{
	// SPI enabled
	// master mode
	// SPIclock = Fosc/16 = 1MHz (less than 20MHz)
  SPCR = (1 << SPE) | (1 << MSTR) | (0 << SPR1) | (1 << SPR0);  // p:134
	sbi(SPSR, SPI2X); // p:134
}

// ---------------------------------------------------------------------------
// Send a char on the SPI and return the answer
// ---------------------------------------------------------------------------
char SPI(char d) 
{  // send character over SPI
	SPDR = d;
	while(!(SPSR & (1<<SPIF)));
	return SPDR;
}

// ---------------------------------------------------------------------------
// Wait a specific value from the SD card. Return true if the expectedValue
// match, false in the other case
// ---------------------------------------------------------------------------
uint8_t sdWait(char expectedValue)
{
	uint16_t ix;
	char r1 = SPI(0xFF);
	for (ix = 0; ix < WAIT_TIMEOUT; ix++)
	{
		if (r1 == expectedValue) 
		break;
		r1 = SPI(0xFF);
	}
	SDCARD_error = r1;
  return (r1 == expectedValue) ;
}

// ---------------------------------------------------------------------------
// Send a command to the SD card
// ---------------------------------------------------------------------------
void sdCommand(char cmd, uint32_t Addr, char crc)
{	// sends a command to the MMC
	SPI(0xFF);
	SPI(0x40 | cmd);
	SPI((uint8_t)(Addr >> 24));
	SPI((uint8_t)(Addr >> 16));
	SPI((uint8_t)(Addr >> 8));
	SPI((uint8_t)Addr);
	SPI(crc);
	SPI(0xFF);
}

// ---------------------------------------------------------------------------
// Macro to deselect the card and return a boolean
// ---------------------------------------------------------------------------
#define SD_RETURN(VALUE)  \
	sdDeSelectChip();  \
	return VALUE;

// ===========================================================================
// PUBLIC FUNCTIONS
// ===========================================================================

// ---------------------------------------------------------------------------
//  This function initializes the SPI ports
// ---------------------------------------------------------------------------
void sdInit_ports()
{
	SDCARD_error = R1_NOERROR;
// init spi port
  cbi(SPI_DDR, SPI_MISO);   // set port B SPI data (MISO) input to input
  sbi(SPI_DDR, SPI_SCK);    // set port B SPI clock to output
  sbi(SPI_DDR, SPI_MOSI);	  // set port B SPI data out (MOSI)to output
  sbi(SPI_DDR, SPI_SS);	    // set port B SPI chip select to output

	sdSetSlowClockSpeed();
  sdDeSelectChip();
} 

// ---------------------------------------------------------------------------
//  Return the last error code
// ---------------------------------------------------------------------------
uint16_t sdGetError()
{
	return SDCARD_error;
}

// ---------------------------------------------------------------------------
// Initialize the SD card. Return true if the card is ready 
// for read/write
// ---------------------------------------------------------------------------
uint8_t sdInitCard(void) 
{ 
  char i;
	uint16_t ix=0;
	sdSetSlowClockSpeed();

// start SPI mode : 80 clocks pulses while sdcard is not selected
  sdDeSelectChip();
  for(i=0; i < 10; i++)
	{
			SPI(0xFF); // send 10*8=80 clock pulses
	}

  sdSelectChip();
// reset the card
	sdCommand(CMD0_GO_IDLE_STATE, 0 ,0x95/*CRC7*/);
	i = SPI(0xFF);
  if (i != R1_IDLE)
  {
		SDCARD_error = i + ERROR_INIT_1;
		SD_RETURN(false);
  }

// wait the end of reset
	ix=0;
	do
	{
		sdCommand(CMD1_SEND_OPCOND, 0, 0xFF/*Dummy CRC*/);
		i = SPI(0xFF);
	} while ((i != R1_NOERROR) && (++ix < WAIT_TIMEOUT));
	
	if (i != R1_NOERROR)
	{
		SDCARD_error = i + ERROR_INIT_2;
		SD_RETURN(false);
	}

// increase clock speed for read/write
	sdSetFastClockSpeed();

  SD_RETURN(true);
}


// ---------------------------------------------------------------------------
// Write a buffer data at the specified block address
// ---------------------------------------------------------------------------
uint8_t sdWriteBlock(sdCardAddr block) 
{ 
// write RAM sector to MMC
	uint8_t c;
	uint16_t i;
	sdSelectChip();
	// 512 byte-write-mode
	sdCommand(CMD24_WRITE_BLOCK, block*BLOCK_SIZE, 0xFF/*dummy crc*/);
	if (!sdWait(R1_NOERROR))
	{
		SDCARD_error += ERROR_WRITE_1;
		SD_RETURN(false);
	}

	SPI(0xFF);
	SPI(0xFF);
	SPI(START_TOKEN);
	// write ram sectors to SDCard
	for (i=0; i < BLOCK_SIZE; i++)
	{
			SPI(buffer[i]);
	}
	// at the end, send 2 dummy bytes (CRC)
	SPI(0xFF);
	SPI(0xFF);

	// get the data response token
	c = SPI(0xFF);
	c &= 0x1F; 	// 0x1F = 0b.0001.1111;
	if ((c>>1) != 0x02)
	{ // 0x02=data accepted ; 0x05=CRC error; 0x06=write error
		SDCARD_error = c;
		SDCARD_error += ERROR_WRITE_2;
		SD_RETURN(false);
	}

	// wait until card is not busy anymore
	if (!sdWait(0xFF))
	{
		SDCARD_error += ERROR_WRITE_3;
		SD_RETURN(false);
	}

	SD_RETURN(true);
}

// ---------------------------------------------------------------------------
// Read data from the SDcard and fill buffer with it
// ---------------------------------------------------------------------------
uint8_t sdReadBlock(sdCardAddr block)
{ 
	uint16_t i;
	sdSelectChip();
	// 512 byte-read-mode 
	sdCommand(CMD17_READ_BLOCK, block*BLOCK_SIZE, 0xFF/*dummy CRC*/);
	if (!sdWait(R1_NOERROR))
	{
		SDCARD_error += ERROR_READ_1;
		SD_RETURN(false);
	}

	// wait for 0xFE - start of any transmission
	if (!sdWait(START_TOKEN))
	{
		SDCARD_error += ERROR_READ_2;
		SD_RETURN(false);
	}

	// proceed with SDCard-read
	for(i=0; i < BLOCK_SIZE; i++)
	{
		buffer[i] = SPI(0xFF); 
	}
	// at the end, send 2 dummy bytes
	SPI(0xFF); // actually this returns the CRC/checksum byte
	SPI(0xFF);

	// wait until card is not busy anymore
	if (!sdWait(0xFF))
	{
		SDCARD_error += ERROR_WRITE_3;
		SD_RETURN(false);
	}

	SD_RETURN(true);
}



/** ***************************************************************************
retourne l'adresse du début des datas dans la SDcard (enregistrés dans un gros fichier de 1Mo
afin d'être accessible par le gestionnaire de fichier sous Linux pour les lire)
avec comme critère de recherche le début de la séquence le texte (16 caractères) : "DEBUT_DATA1_ICI:"
Cette façon rudimentaire de procéder permet de s'affranchier de la gestion des FAT, FILE SYSTEM et des fichiers sur la SDcard
tout en traitant des données lisibles par le système Linux, puisque situées dans un fichier, et non directement dans la mémoire hors de toute partition.

****************************************************************************/
uint32_t get_debut_data()
{
	uint8_t etiquette[16+1] = "DEBUT_DATA1_ICI:";
	uint8_t str_lu[40];
	uint32_t num_block;
	uint32_t compteur1;
	uint8_t i;
	uint16_t k;
	uint32_t j, j_max;
	uint8_t ok;
	
	num_block = 1;
	compteur1=0;
	j_max = 2e6 ;  // 1e9 / 512

	ok=0;

	for(j=0; j<j_max; j++)
	{
		sdReadBlock(j);
		for(k=0; k<512; k++)
		{
			//décallage à gauche du str_lu
			for (i=0; i<40; i++) {str_lu[i] = str_lu[i+1]; }
			str_lu[39]=buffer[k];
			//compare buffer[0..5] avec l'etiquette
			ok=1;
			for (i=0; i<6; i++) 
			{ 
				if (str_lu[i] != etiquette[i]) { ok=0; }	
			}
			if (ok==1) return (compteur1);	//sortir si comparaison ok

			compteur1++;
			
			//lcd_gotoxy(10, 2);
			//lcd_putc(65 + ((compteur1 / 2000 ) % 26)); // tit' animation... pour patienter
			//if (byteCounter >= fileSize ) return 1;// (séquence pas trouvée)
		}
		lcd_goto_LC(2, 10);
		lcd_aff_nb (j, 6, 0);
	}
		
	
	return compteur1;
}



// ===========================================================================
// PUBLIC REGISTER FUNCTIONS
// ===========================================================================

#ifdef CARD_REGISTER_COMMANDS

// ---------------------------------------------------------------------------
// Retrieve the 'responseLength' bytes from the SD card after having sent 
// 'command'
// Output: buffer[responseLength]
// ---------------------------------------------------------------------------
uint8_t sdGetRegister(uint8_t command, unsigned char* buffer,  uint8_t responseLength)
{
	uint8_t i;
	sdSelectChip();

// send the command
	sdCommand(command ,0, 0xFF/*dummy CRC*/);
	if (!sdWait(R1_NOERROR))
	{
		SDCARD_error += ERROR_REGISTER_1;
		SD_RETURN(false);
	}

// wait for 0xFE - start of any transmission
	if (!sdWait(START_TOKEN))
	{
		SDCARD_error += ERROR_REGISTER_2;
		SD_RETURN(false);
	}

	for(i=0; i < responseLength; i++)
	{
		buffer[i] = SPI(0xFF); 
	} 

	// CRC 16
	SPI(0xFF); 
	SPI(0xFF); 

	// wait until card is not busy anymore
	if (!sdWait(0xFF))
	{
		SDCARD_error += ERROR_REGISTER_3;
		SD_RETURN(false);
	}

	SD_RETURN(true);
}


// ---------------------------------------------------------------------------
// Retrieve the 16 bytes Card Identification register
// ---------------------------------------------------------------------------
uint8_t sdGetCSD(unsigned char* buffer)
{
	return sdGetRegister(CMD9_SEND_CSD, buffer, 16);
}



// ---------------------------------------------------------------------------
// Retrieve the 16 bytes Card Identification register
// ---------------------------------------------------------------------------
uint8_t sdGetCID(unsigned char* buffer)
{
	return sdGetRegister(CMD10_SEND_CID, buffer, 16);
}



// ---------------------------------------------------------------------------
//  Retrieve the 4 bytes of the OCR register
// ---------------------------------------------------------------------------
uint8_t sdGetOCR(unsigned char* buffer)
{
	uint8_t i, r1;
  sdSelectChip();
// 512 byte-read-mode 
	sdCommand(CMD58_READ_OCR ,0, 0xFF/*dummy CRC*/);
	r1 = SPI(0xFF);

	for(i=0; i < 4; i++)
	{
			buffer[i] = SPI(0xFF); 
	}
    
	if ( r1 != R1_NOERROR) {
		SDCARD_error = r1 + ERROR_OCR_1;
        SD_RETURN(false);
	}

  SD_RETURN(true);
}

// ---------------------------------------------------------------------------
// Retrieve the 2 bytes of the status register
// ---------------------------------------------------------------------------
uint8_t sdGetStatus(unsigned char* buffer)
{
    sdSelectChip();
    // 512 byte-read-mode 
    sdCommand(CMD13_SEND_STATUS ,0, 0xFF/*dummy CRC*/);
    buffer[0] = SPI(0xFF); 
    buffer[1] = SPI(0xFF); 

    SD_RETURN(true);
}

#endif // CARD_REGISTER_COMMANDS
